//
// Created by nicolas on 20/03/18.
//

#include <stddef.h>
#include <stdlib.h>
#include <assert.h>
#include "queue.h"
#include "mutex.h"

struct st_node {
    void* item;
    struct st_node* next;
};
struct st_queue {
    mutex_t lock;
    size_t size;
    struct st_node* start;
    struct st_node* end;
};

queue_t queue_create() {
    queue_t q = malloc(sizeof(struct st_queue));
    assert(q != 0);
    q->lock = mutex_create();
    q->size = 0;
    q->start = 0;
    q->end = 0;
    return q;
}

void queue_enqueue(queue_t q, void* item) {
    struct st_node* node = malloc(sizeof(struct st_node));
    assert(node != 0);
    node->item = item;
    node->next = 0;

    mutex_lock(q->lock);
    if (q->end) {
        q->end->next = node;
    }
    if (q->start == 0) {
        q->start = node;
    }
    q->end = node;
    q->size++;
    mutex_unlock(q->lock);
}

void* queue_dequeue(queue_t q) {
    mutex_lock(q->lock);
    if (q->start == 0) {
        mutex_unlock(q->lock);
        return 0;
    }
    struct st_node* node = q->start;
    void* item = node->item;
    q->start = node->next;
    q->size--;
    mutex_unlock(q->lock);
    free(node);
    return item;
}

void* queue_peek(queue_t q) {
    mutex_lock(q->lock);
    void* item = q->start ? q->start->item : 0;
    mutex_unlock(q->lock);
    return item;
}

inline size_t queue_size(queue_t q) {
    mutex_lock(q->lock);
    size_t size = q->size;
    mutex_unlock(q->lock);
    return size;
}

void queue_destroy(queue_t q) {
    mutex_lock(q->lock);
    struct st_node* p = q->start;
    while(p) {
        struct st_node* next = p->next;
        free(p);
        p = next;
    }
    mutex_unlock(q->lock);
    free(q);
}
