//
// Created by nicol on 2018-03-18.
//

#include <stdlib.h>
#include "thread.h"

#ifdef _WIN32
#include <windows.h>

struct st_thread {
    DWORD   id;
    HANDLE  handle;
    thread_start_t start;
    thread_arg_t arg;
    int     last_error;
};

thread_t thread_create(thread_start_t thread_start, thread_arg_t thread_arg) {
    thread_t thread = malloc(sizeof(struct st_thread));
    thread->handle = 0;
    thread->start = thread_start;
    thread->arg = thread_arg;
    thread->last_error = 0;
    return thread;
}

int thread_start(thread_t thread) {
    thread->handle = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)thread->start, thread->arg, 0, &thread->id);
    if (thread->handle == NULL) {
        DWORD rc = GetLastError();
        thread->last_error = (int)rc;
        return (int)rc;
    }
    return 0;
}

int thread_wait(thread_t thread) {
    if (WaitForSingleObject(thread->handle, INFINITE) == WAIT_FAILED) {
        DWORD rc = GetLastError();
        thread->last_error = (int)rc;
        return (int)rc;
    }
    return 0;
}

int thread_destroy(thread_t thread) {
    if (thread_wait(thread) != 0) SuspendThread(thread->handle);
    CloseHandle(thread->handle);
    free(thread);
}

int thread_error(thread_t thread) {
    return thread->last_error;
}

#endif
#ifdef __linux__
#include <pthread.h>

struct st_thread {
    pthread_t handle;
    thread_start_t start;
    thread_arg_t arg;
    int last_error;
    void* result;
    int joined;
};

thread_t thread_create(thread_start_t thread_start, thread_arg_t thread_arg) {
    thread_t thread = malloc(sizeof(struct st_thread));
    thread->handle = 0;
    thread->start = thread_start;
    thread->arg = thread_arg;
    thread->last_error = 0;
    thread->result = 0;
    thread->joined = 0;
    return thread;
}

int thread_start(thread_t thread) {
    int rc = pthread_create(&thread->handle, NULL, thread->start, thread->arg);
    if (rc != 0) {
        thread->last_error = rc;
        return -1;
    }
    return 0;
}

int thread_wait(thread_t thread) {
    if (thread->joined) {
        return 0;
    }
    void* result;
    int rc = pthread_join(thread->handle, &result);
    if (rc != 0) {
        thread->last_error = rc;
        return -1;
    }
    thread->joined = 1;
    return 0;
}

int thread_destroy(thread_t thread) {
    if (thread_wait(thread)) {
        int rc = pthread_cancel(thread->handle);
        if (rc != 0) {
            thread->last_error = rc;
            return -1;
        }
    }
    free(thread);
    return 0;
}

inline int thread_error(thread_t thread) {
    return thread->last_error;
}

inline int thread_result(thread_t thread, void** result) {
    if (thread_wait(thread)) {
        return -1;
    }
    *result = thread->result;
    return 0;
}
#endif
